module Engine.Relational where

import AST.Vector
import AST.Prolog
import AST.Relational
import Engine.Rewritable
import Engine.Unification (unify, apply)

import Compiler.Module

import Data.List
import Control.Monad.Trans
import Control.Monad.State

type RModule = Module Predicate Relational 

instance (MonadIO m, MonadState RModule m) => Rewritable m Relational where
    isFinal = relIsFinal
    rewriteStep = relRewrite

-- This module is at an early stage of implementation.

debug s = liftIO $ putStr s

-- See AST.Relational about invariants that hold here.

relIsFinal :: Monad m => Relational -> m Bool
{-- An union is final if any member of it it's too.
relIsFinal (CUnion []) = return True
relIsFinal (CUnion l) = 
    do lbool <- mapM (relIsFinal) l 
       return $ foldl (||) False lbool
-- For an intersection all elems must be final.
-- egallego@babel: Ummm, maybe an intersection is never final.
relIsFinal (CIntersection l) = 
    do lbool <- mapM (relIsFinal) l
       return $ foldl (&&) True lbool
-- A concat: Have to think.
relIsFinal (CConcat l)   =  
    do lbool <- mapM (relIsFinal) l
       return $ foldl (&&) True lbool
-- Every function is final except a Predicate.
relIsFinal (FPred s) = return False -}
relIsFinal (FK _   ) = return True
relIsFinal _         = return False

-- We filter any list of a single elem
--filterl [x] = x
--filterl l   = l

-- Here we model the rewrite rules as pattern matching in Haskell.

relRewrite                   :: (MonadIO m, MonadState RModule m) => Relational -> m Relational
relRewrite (CUnion l)        = rewriteUnion l
relRewrite (CIntersection l) = rewriteInter l
relRewrite (CConcat l)       = rewriteConcat l
relRewrite (FPred p )        = 
    do m <- get 
       case lookupCode m p of
	    (newexpr:_) -> return newexpr
	    []          -> return F0
			       
relRewrite x                 = return x

-- First we try to rewrite the known cases, then we continue.

rewriteConcat []  = return $ CConcat []
rewriteConcat [x] = return x
rewriteConcat [FI n, FK v, FI n'] = 
    let v' = take n v in
	 return $ FK v'

rewriteConcat [FConv (FW po), FK v, FW o] = 
    let newk = permute v (invert o) in
	do --debug $ "\nPermute \n ko: " ++ (show v) ++ "\n\nkn: " ++ (show newk) ++ "\n\nConv : " ++ (show o) ++"\n\n"
	  return $ FK newk

rewriteConcat [FW po, FK v, FConv(FW o)] = 
    let newk = permute v po in
	   return $ FK newk

rewriteConcat [FW o, CUnion l, FConv(FW po)] =
    return $ CUnion $ map (\e -> (CConcat [FW o, e, FConv(FW po)])) l

rewriteConcat [FI n, CUnion l, FI n'] =
    return $ CUnion $ map (\e -> (CConcat [FI n, e, FI n'])) l

rewriteConcat l = 
    if elem F0 l then return F0
    else
    do rl <- mapM relRewrite l
       return $ CConcat rl

-- This belongs in IR.Vector
invertPerm :: Perm -> Perm
invertPerm (CPerm p1 p2) = CPerm p2 p1


rewriteUnion []     = return $ F0
-- Kill 0
rewriteUnion (F0:r) = rewriteUnion r
rewriteUnion (x:[]) = 
    relRewrite x
rewriteUnion (x:xs) =  
    do --debug $ "rewriteUnion called\n " ++ (show x) ++ "\n"
       newx  <- relRewrite x
       final <- relIsFinal newx 
       case final of 
		True  -> do newxs <- rewriteUnion xs
		 --	    debug $ "rewriteUnion: rewriting tail\n" ++ (show newxs) ++ "\n"
			    case newxs of 
			         (CUnion l) -> return $ CUnion (newx:l)
				 F0         -> return newx
				 e          -> return $ CUnion [newx,e]
		False -> return $ CUnion (newx:xs)
-- Logic reductions.

rewriteInter []  = return $ CIntersection []
rewriteInter [x] = do --debug "aqui estamos \n"
		      return x

rewriteInter [FK v1,FK v2] = 
    let u = tryUnify v1 v2 in
	    case u of 
		   Right v  -> return $ (FK v)
		   Left err -> return $ F0

rewriteInter (FK v1:FK v2:r) = 
    let u = tryUnify v1 v2 in
	    case u of 
		   Right v  -> return $ CIntersection $ (FK v):r
		   Left err -> return $ CIntersection $ F0:r


-- (l1 U l2 U l3) /\ C -> (l1 /\ C) U (l2 /\ C) ...
rewriteInter [(CUnion l), c] = 
    return $ CUnion (map (\le -> CIntersection [c, le]) l)

-- C /\ (l1 U l2 U l3) -> (l1 /\ C) U (l2 /\ C) ...
rewriteInter [c, (CUnion l)] =
    return $ CUnion (map (\le -> CIntersection [c, le]) l)

-- A /\ (B /\ C) -> (A /\ B /\ C) 
rewriteInter [a, (CIntersection l)] =
    return $ CIntersection $ (a:l)

-- Comment this meta redo out 
-- Meta-reductions.

rewriteInter [a, CConcat [FI n, e@(CIntersection li), FI n']] =
--rewriteInter [a, CConcat [FI n, e, FI n']] =
    return $ CIntersection $
	       [ CConcat [ FI n, 
			   CIntersection 
			     [ CConcat [ FI n, a, FI n ], e ],
			   FI n
			 ], a
	       ]

{-rewriteInter [CConcat [FI n, e@(CIntersection li), FI n'],a] =
    return $ CIntersection $
	       [ CConcat [ FI n, 
			   CIntersection 
			     [ CConcat [ FI n, a, FI n ], e ],
			   FI n
			 ], a
	       ]
-}

-- Basic rewriting rule.
rewriteInter [a, CConcat [FW p, q, FConv (FW po)]] =
    return $ CConcat 
	       [ FW p, CIntersection 
		 [ CConcat 
		   [ FConv (FW po), 
		     a, 
		     FW p
		   ], 
		   q
		 ], 
		 FConv (FW po)
	       ]

-- Basic rewriting rule.
rewriteInter (a: CConcat [FW p, q, FConv (FW po)]:r) =
    return $ CIntersection $ 
	       [ CConcat 
		 [ FW p, CIntersection 
		   [ q, 
		     CConcat 
		     [ FConv (FW po), 
		       a, 
		       FW p
		     ]		     
		   ], 
		   FConv (FW po)
		 ]] ++ r

   
rewriteInter ll@(l:ls) =
    if elem F0 ll then
       return F0
    else
{-       do newl  <- relRewrite l
	  final <- isFinal newl
	  if final then 
	     do newls <- rewriteInter ls 
		return $ CIntersection (newl:newls:[])
             else  
	     do newls <- rewriteInter ls 
		return $ CIntersection (newl:newls:[])-}
    do nl <- mapM relRewrite ll --liftIO $ putStr $ "No pattern found" ++ (show l)
       return $ CIntersection nl

-- By cal
permute :: Vector -> Perm -> Vector
permute [] (CPerm [] p) = p
permute v  (CPerm p [])       = []
permute v (CPerm p1 p2)       = 
    let v' = zip p2 v 
	v''= sortBy (\(n1,_) (n2,_) -> compare n1 n2) v'
	(_,nv) = unzip v'' in
	nv

invert               :: Perm -> Perm
invert (CPerm [] p)  = CPerm p []
invert (CPerm p [])  = CPerm [] p
invert (CPerm p1 p2) = 
    let np  = zip p2 p1
	np' = sortBy (\(n1,_) (n2,_) -> compare n1 n2) np
	(np1, np2) = unzip np' in
	CPerm np1 np2

-- Vector to term.
-- This converts a vector of terms in a fake term
-- to allow unification.
vat n v = CTStruct "fake" (freezeCTVars n v)
tav (CTStruct _ v) = v
tav x = [x]

freezeCTVars :: Int -> Vector -> Vector
freezeCTVars n l = map (freezeTerm n) l

freezeTerm                 :: Int ->Term -> Term
freezeTerm i (CTFreeVar n)   = CTVar ("_" ++(show i)++"x"++(show n))
freezeTerm i (CTStruct n lt) = CTStruct n $ map (freezeTerm i) lt
freezeTerm i t = t

tryUnify :: Vector -> Vector -> Either String Vector
tryUnify v1 v2 = 
    let minl = min (length v1) (length v2)
	v1'  = take minl v1
	v2'  = take minl v2 
	(nmax, vmax) = if (length v1) > (length v2) then (1, v1) else (2, v2) in
	  do s <- unify (vat 1 v1') (vat 2 v2')
	     return $ tav (apply s (vat nmax vmax))

