module Translator.Prolog.Relational where

import Util.StrictMap

import AST.Prolog
import AST.Vector
import AST.Relational

import Compiler.Module (Module, mapModuleList)

import Translator 

import Data.List      ( (\\) )
import Control.Monad.State

-- Utils for Relational translation

type TransM a   = State TransState a

-- With this data type we track each mapping made when introducing new
-- fresh vars.
type VarMapping = StrictMap Term Term

data TransState = TS 
    { 
     varmap  :: VarMapping,
     used    :: Int 
    }

initialTransState = 
    TS {
	varmap   = emptySM,
	used     = 0
       }

newMapping :: Term -> Term -> TransM ()
newMapping t1 t2 = do
    ts <- get
    put $ ts { varmap = (addToSM (varmap ts) t1 t2) }  
    return ()

--
lookupBothSM :: Ord a => StrictMap a a -> a -> Maybe a
lookupBothSM sm o =
    case lookupFirstSM sm o of
         Just v  -> Just v
	 Nothing -> lookupSecondSM sm o


hasMapping :: Term -> TransM (Maybe Term)
hasMapping t1 = do
    ts <- get
    return $ lookupBothSM (varmap ts) t1
  

mkNewVar   :: TransM Term
mkNewVar = do 
    ts <- get
    put $ ts { used = ((used ts) + 1) }
    return $ CTFreeVar ((used ts)+1)
  
----------------------------------------------------------------
-- Normalizes a predicate.

normalizePredicate :: (Predicate, [Clause]) -> (Predicate, [Clause])
normalizePredicate (p, l) = (p, map fnorm l) 
    where
	fnorm x = let (nc, _) = normalizeClause x in
				nc


-- Returns normalized clauses.

normalizeClause :: Clause -> (Clause, VarMapping) 
normalizeClause c = 
    let nc            = normalizeClause' c
	(nc2, state)  = runState nc initialTransState 
	maps          = varmap state  in
--	fmap'         = renameBoundSM mappings        in
			(nc2, maps)


normalizeClause' :: Clause -> TransM Clause
normalizeClause' (CClause head tail) = do
    head' <- normalizeHead head
    tail' <- normalizeTail tail
    return $ CClause head' tail'
    


----------------------------------------------------------------
-- Similar to normalizeAtom except always susbtitute with a fresh var

normalizeHead :: Atom -> TransM Atom
normalizeHead (CAtom p l) = do
    nl1 <- mapM replaceTermWithFresh l
    nl2 <- mapM renameBoundVars l
    return $ CAtom p nl1

normalizeTail :: [Atom] -> TransM [Atom]
normalizeTail l = do
    na <- mapM normalizeAtom l
    return na

-- This function extracts all of the terms in the atom and replaces
-- them with fresh variables.

normalizeAtom :: Atom -> TransM Atom
normalizeAtom (CAtom p l) = do
    nl <- mapM normalizeTerm l 
    return $ CAtom p nl

----------------------------------------------------------------------------------------
-- FIXME: The three functions below replaceTermWithFresh, renameBoundVars, normalizeTerm
-- should be merged and rethinked.

-- We replace every term with a fresh var.
-- If we're normalizing add(o, X, X), for example, we
-- must check if the term is a bound var if it's has
-- a previus mapping and make the new mapping (in this case):
-- x3 -> x2 instead of x3->X again.
replaceTermWithFresh :: Term -> TransM Term
replaceTermWithFresh t@(CTVar v) = do
    nv <- mkNewVar
    tm <- hasMapping t
    case tm of
	    Just nt -> newMapping nv nt
	    Nothing -> newMapping nv t
    return nv

replaceTermWithFresh t = do
    nv <- mkNewVar
    newMapping nv t
    return nv


-- Rename bound vars.
renameBoundVars :: Term -> TransM Term
renameBoundVars t@(CTVar v) = do
    tm <- hasMapping t
    case tm of 
	    Just nt -> return nt
	    Nothing -> do nv <- mkNewVar
			  newMapping nv t
			  return nv
renameBoundVars t@(CTStruct n lt) = do
    tm <- hasMapping t
    case tm of
	    Just nt -> do nlt <- mapM renameBoundVars lt
			  newMapping nt (CTStruct n nlt)
			  return (CTStruct n nlt)
	    Nothing -> do nlt <- mapM renameBoundVars lt
			  return (CTStruct n nlt)

renameBoundVars t = return t


-- Substitutes every bound Var or mapped term with their correspoding
-- free var.
normalizeTerm                   :: Term -> TransM Term
normalizeTerm t@(CTFreeVar _)   = return t
-- Don't make new mappings of constants :)
normalizeTerm t@(CTConst   v)   = return t
normalizeTerm t@(CTVar     v)   = do
    m <- hasMapping t  
    case m of  
	   Just mp   -> return mp  
	   Nothing   -> do nv <- mkNewVar  
			   newMapping nv t
			   return nv  
-- Here we recursively normalize terms. If the term is a struct, we 
-- must see if it has a mapping. When we have a mapping for the struct 
-- we normalize it's inner terms and create a new mapping.
normalizeTerm t@(CTStruct n lt) = do 
    m <- hasMapping t
    case m of  
   	   Just mp   -> return mp

	   Nothing   -> do nv  <- mkNewVar  
			   ntl <- mapM renameBoundVars lt
			   newMapping nv (CTStruct n ntl)
			   return nv  
								    
-----------------------------------------------------------

--arity :: PrologProgram -> Int
--arity fm = case  

----

instance Translator (Module Predicate Clause) Predicate Relational where
    translate a = Right $ translateProlog a

translateProlog :: Module Predicate Clause -> 
		   Module Predicate Relational
translateProlog pprog = 
    mapModuleList translatePredicate pprog
	

translatePredicate :: Predicate -> [Clause] -> [Relational]
translatePredicate p l =
    let np = map normalizeClause l in
	     [CUnion $ map translateClause np]
	

-- Head is now normalized.
translateClause :: (Clause, VarMapping) -> Relational
translateClause ((CClause head lt), maps) = 
    let innerexpr = (translateMappings maps) : (map (translateAtom maps) lt)
	arity'    = FI (arity head) in
	     case innerexpr of
	          -- If innerexpr is a single elem, don't intersect it.
	          [x] -> CConcat $ [arity', x, arity']
		  _   -> CConcat $ [arity', CIntersection innerexpr, arity']

-- Each atom is translated to a W(p) ; atom ; W(p)
translateAtom :: VarMapping -> Atom -> Relational
translateAtom maps a@(CAtom n lt)  = 
    let perm = constructPerm maps a 
	pred = (n, length lt)          in
	       CConcat [FW perm, FPred pred, FConv (FW perm)]

-- Here we are going to get the K relation.
-- We have mapping along x1 -> s(o), x2 -> X, so if we have a fresh var mapping to
-- a not fresh var we should name in K the fresh, that is the var the relational
-- machine will know.
translateMappings :: VarMapping -> Relational
translateMappings sm = 
    let lt              = smToList sm      
	extract_mapping = \(fresh, term) -> 
			    case term of (CTVar x) -> fresh
					 _         -> term      in
	FK (map extract_mapping lt)


-- This constructs the permutation 
constructPerm :: VarMapping -> Atom -> Perm
constructPerm sm (CAtom n lt) = 
    let lm  = map (\(t,_)->t) (smToList sm)
	lt' = fillPerm lm lt                 in
	     CPerm lm lt'

-- Fill perm includes dummy variables to fill the vector with n vars
-- Slow algorithm.
fillPerm :: Vector -> Vector -> Vector
fillPerm lm lt = 
    let diff = lm \\ lt in
	       lt ++ diff
    
-- This belongs in IR.Vector
invertPerm :: Perm -> Perm
invertPerm (CPerm p1 p2) = CPerm p2 p1
